/* tmndec.c, main(), initialization, option processing                    */

/*
 * tmndecode 
 * Copyright (C) 1995 Telenor R&D
 *                    Karl Olav Lillevold <kol@nta.no>                    
 *
 * based on mpeg2decode, (C) 1994, MPEG Software Simulation Group
 * and mpeg2play, (C) 1994 Stefan Eckart
 *                         <stefan@lis.e-technik.tu-muenchen.de>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>

#define GLOBAL
#include "config.h"
#include "tmndec.h"
#include "global.h"
#define IMPLEMENT_Pl
#include "Pl.h"

#ifdef SH_MEM

#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>


/* private prototypes */
static void display_image _ANSI_ARGS_((XImage *ximage, unsigned char *dithered_image));



extern int XShmQueryExtension _ANSI_ARGS_((Display *dpy));
extern int XShmGetEventBase _ANSI_ARGS_((Display *dpy));

static int HandleXError _ANSI_ARGS_((Display *dpy, XErrorEvent *event));
static void InstallXErrorHandler _ANSI_ARGS_((void));
static void DeInstallXErrorHandler _ANSI_ARGS_((void));

static int shmem_flag;
static XShmSegmentInfo shminfo1, shminfo2;
static int gXErrorFlag;
static int CompletionType = -1;
#endif

/* DISPLAY */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>
#include <Xm/Label.h>

/* PLUGIN */

#include "npapi.h"

/* END PLUGIN */

#ifdef USE_TIME
#ifndef WIN32
#include <sys/time.h>
#else
#include <windows.h>
#endif
#endif
#ifdef WIN32
#include <io.h>
#endif

/* DISPLAY */

/* display related data */
unsigned long wpixel[3];
static unsigned char *dithered_image;

/* X11 related variables */
static Display *display;
static Window window, okno_from_netscape;
static Visual *visual;
static GC gc;
static Colormap cmap;
static XEvent xev;
static int dpy_depth;
int newcmap;

XWindowAttributes xwa;
unsigned long bg, fg;
int NW_width, NW_height;

XImage *ximage;

char *inputname;
unsigned char pixel[256];
XColor c[209]; /* will keep RGB coeffs for video colormap */

void start();
void my_pause();
void replay();
void toggled();
void toggformat();
void SetMessage(const char *msg);
Widget messagelabel;

typedef struct _PluginInstance
{
    Widget	netscape_widget; /* The widget provided by netscape */
    Display    *display;         /* Netscape's display connection */
    uint16	mode;            /* NP_EMBED, NP_FULL, or NP_BACKGROUND */
    uint32 width, height;

    /*--- Start custom code */
    Window inWindow;
    GC gc;
    XtAppContext appContext;
    int winLines;
    int winPels;
    int buttonsLine;
/*--- End custom code */
} PluginInstance;

JRIEnv* env;
PluginInstance *This;
NPP plugin, npp;
NPStream *strumien, *oldstrumien;
XmString label;
Widget button1, button2, button3, tbutton, tformat;

int curwrt;
int local_flag=0;
int started=0;
XtIntervalId timerId=0;

#ifdef USE_TIME
  int runtime;
#ifndef WIN32
  struct timeval tstart,tstop, pstart, pstop, ptime;
#else
  unsigned int startTime, stopTime;
#endif
#endif

/* END PLUGIN */

#ifdef WINDOWS
int initDisplay (int pels, int lines);
int closeDisplay ();
#endif

/* private prototypes */
static int finishdecoder _ANSI_ARGS_((void));
static int initdecoder _ANSI_ARGS_((void));
static void options _ANSI_ARGS_((int *argcp, char **argvp[]));
static int getval _ANSI_ARGS_((char *argv[]));
int decode_data();

/* PLUGIN */
/* **************** Here starts code for plugin in Netscape********** */
char *
NPP_GetMIMEDescription(void)
{
  return("video/h263:263:Video Player");
}


NPError
NPP_GetValue(void *future, NPPVariable variable, void *value)
{
	NPError err = NPERR_NO_ERROR;

	switch (variable) {
		case NPPVpluginNameString:
			*((char **)value) = "H.263 player plug-in";
			break;
		case NPPVpluginDescriptionString:
			*((char **)value) =
				"This plugins handles h.263 video files and plays "
				 "them in a window with some controls.";
			break;
		default:
			err = NPERR_GENERIC_ERROR;
	}
	return err;
}

NPError
NPP_Initialize(void)
{
  return NPERR_NO_ERROR;
}

jref
NPP_GetJavaClass()
{
   env = NPN_GetJavaEnv();
   return use_Pl(env);
}

void
NPP_Shutdown(void)
{
                unuse_Pl(env);
}

NPError 
NPP_New(NPMIMEType pluginType,
	NPP instance,
	uint16 mode,
	int16 argc,
	char* argn[],
	char* argv[],
	NPSavedData* saved)
{

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
	
	This = (PluginInstance*) instance->pdata;
	
	playmode = WAIT;
	started = 0;

	if (This != NULL)
	{
   	        int i;
		This->netscape_widget = NULL;
		This->display = NULL;
		This->mode = mode;    /* mode is NP_EMBED, NP_FULL, or NP_BACKGROUND (see npapi.h) */
		This->width = 0;
		This->height = 0;
		This->inWindow = NULL;
		if (mode==NP_FULL)
		{
                  This->winPels = 500;
                  This->winLines = 370;
		}
		else
		{
                  This->winPels = 176;
                  This->winLines = 144;
		}
		This->buttonsLine = 45;
  	        expand=1;

		/* *Developers*: Initialize fields of your plugin
		 * instance data here.  If the NPSavedData is non-
		 * NULL, you can use that data (returned by you from
		 * NPP_Destroy to set up the new plugin instance.
		 */
                /* get any args we recognize from the html file */

		if (mode==NP_EMBED)
		{
                for(i = 0; i < argc; i++)
                {
                    if(!strcmp(argn[i], "FORMAT"))
                    {
                        if(!strcmp(argv[i], "QCIF"))
			{
                            This->winPels = 176;
                            This->winLines = 144;
			    expand=0;
			  }
                        if(!strcmp(argv[i], "CIF"))
			{
                            This->winPels = 352;
                            This->winLines = 288;
			    expand=1;
			  }
                    }
		  }
	      } /* if EMBED */
		return NPERR_NO_ERROR;
	}
	else
	{
		return NPERR_OUT_OF_MEMORY_ERROR;
      }
}/* NPP_New */


NPError 
NPP_Destroy(NPP instance, NPSavedData** save)
{

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;

	/* PLUGIN DEVELOPERS:
	 *	If desired, call NP_MemAlloc to create a
	 *	NPSavedDate structure containing any state information
	 *	that you want restored if this plugin instance is later
	 *	recreated.
	 */

	if (This != NULL) {
	  
		if (newcmap) XUninstallColormap(display, cmap);
		finishdecoder();
		NPN_MemFree(instance->pdata);
		instance->pdata = NULL;
	}

	fflush(stdout);
	return NPERR_NO_ERROR;
}/* NPP_Destroy */



NPError 
NPP_SetWindow(NPP instance, NPWindow* window)
{
	XGCValues dane;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;
if (This->inWindow == (Window)window->window)
{
  /* subsequent call to SetWindow() - resizing? */
}
else
{
	/*
	 * PLUGIN DEVELOPERS:
	 *	Before setting window to point to the
	 *	new window, you may wish to compare the new window
	 *	info to the previous window (if any) to note window
	 *	size changes, etc.
	 */

	This->display =((NPSetWindowCallbackStruct *)window->ws_info)->display;
	This->inWindow = (Window)window->window;
	This->width = window->width;
	This->height = window->height;
	
	display=This->display;
	
	This->netscape_widget = XtWindowToWidget(This->display, (Window)window->window);
	fflush(stdout);
	
	/*--- Start custom code */
	
        okno_from_netscape=XtWindow(This->netscape_widget);

	XGetWindowAttributes(display, okno_from_netscape, &xwa);

	dpy_depth=xwa.depth;
	visual=xwa.visual;
	gc=(xwa.screen)->default_gc;
	XtVaGetValues(This->netscape_widget, XmNbackground, &dane.background,
		      XmNforeground, &dane.foreground, 0);
	gc = XCreateGC(display, okno_from_netscape, GCForeground | GCBackground, &dane);
	bg=dane.background;
	fg=dane.foreground;
	
	This->gc=gc;
	
	This->appContext = XtWidgetToApplicationContext(This->netscape_widget);

	label = XmStringCreateSimple("PLAY");
	button1 = XtVaCreateManagedWidget("play", xmPushButtonWidgetClass,
					 This->netscape_widget,
					 XmNlabelString, label, 
					 XmNx, 34, 
					 XmNy, This->buttonsLine,
					 XmNwidth, 56,
					 XmNheight, 25,
					 NULL);

	XmStringFree(label);
	XtAddCallback(button1, XmNactivateCallback, start, NULL);

	label = XmStringCreateSimple("PAUSE");
	button2 = XtVaCreateManagedWidget("pause", xmPushButtonWidgetClass,
					 This->netscape_widget,
					 XmNlabelString, label, 
					 XmNx, 100, 
					 XmNy, This->buttonsLine,
					 XmNwidth, 56,
					 XmNheight, 25,
					 NULL);

	XmStringFree(label);
	XtAddCallback(button2, XmNactivateCallback, my_pause, NULL);

	label = XmStringCreateSimple("RePlay");
	button3 = XtVaCreateManagedWidget("replay", xmPushButtonWidgetClass,
					 This->netscape_widget,
					 XmNlabelString, label, 
					 XmNx, 166, 
					 XmNy, This->buttonsLine,
					 XmNwidth, 56,
					 XmNheight, 25,
					 NULL);

	XmStringFree(label);
	XtAddCallback(button3, XmNactivateCallback, replay, NULL);
	XtSetSensitive(button3, False);

	label = XmStringCreateSimple("Do not dump to file");
	tbutton = XtVaCreateManagedWidget("dump",
					 xmToggleButtonWidgetClass,
					 This->netscape_widget,
					 XmNlabelString, label, 
					 XmNx, 34, 
					 XmNy, This->buttonsLine+30,
					 XmNwidth, 146,
					 XmNheight, 25,
					 NULL);

	XmStringFree(label);
	XtAddCallback(tbutton, XmNvalueChangedCallback, toggled, NULL);

	label = XmStringCreateSimple("Expand to CIF");
	tformat = XtVaCreateManagedWidget("expand",
					 xmToggleButtonWidgetClass,
					 This->netscape_widget,
					 XmNlabelString, label, 
					 XmNx, 300, 
					 XmNy, This->buttonsLine+30,
					 XmNwidth, 110,
					 XmNheight, 25,
					 NULL);

	XmStringFree(label);
	XtAddCallback(tformat, XmNvalueChangedCallback, toggformat, NULL);

	/* Creation of messagelabel */
	messagelabel = XtVaCreateManagedWidget( "messagelabel",
			xmLabelWidgetClass,
			This->netscape_widget,
			XmNx, 20,
			XmNy, 10,
			XmNwidth, 300,
			XmNheight, 10,
			NULL );
	SetMessage("Press PLAY to start...");

	XtRealizeWidget(This->netscape_widget);
	/*--- End custom code */
      }
	return NPERR_NO_ERROR;

}/* NPP_SetWindow */


NPError 
NPP_NewStream(NPP instance,
	      NPMIMEType type,
	      NPStream *stream, 
	      NPBool seekable,
	      uint16 *stype)
{

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;

/*--- Start custom code */
	/* Initial Netscape window size set to QCIF */

	NW_width=242;
	NW_height=255;
	/* We want a stream of data */

	*stype = NP_NORMAL;

	/* initialize variables */
	    
	newcmap=0;
	dump_flag=0;
	local_flag=0;
	if (!strncmp(stream->url, "file:",5)) 
	{
	  inputname = (char*)malloc(strlen(stream->url)-4);
	  strncpy(inputname, stream->url+5, strlen(stream->url)-5);
	  strcat(inputname,"\0");
	  local_flag=1;
	}

	plugin=instance;
	strumien=malloc(sizeof(NPStream));
	memcpy(strumien, stream, sizeof(NPStream));

	outtype=4;
	first=1;
	quiet=1;
	spoiler=(char*)malloc(SLOTS*VIDEOUNIT); /* three data slots VIDEOUNIT B each */
	tempbuf=(char*)malloc(SLOTS*VIDEOUNIT);	/* auxiliary buffer */

	bank=SLOTS;
	required=bank*VIDEOUNIT;
	loaded=0;
	curwrt=0;

	if (local_flag) decode_data();	
/*--- End custom code */
	
	return NPERR_NO_ERROR;
}/* NPP_NewStream */


/* PLUGIN DEVELOPERS:
 *	These next 2 functions are directly relevant in a plug-in which
 *	handles the data in a streaming manner. If you want zero bytes
 *	because no buffer space is YET available, return 0. As long as
 *	the stream has not been written to the plugin, Navigator will
 *	continue trying to send bytes.  If the plugin doesn't want them,
 *	just return some large number from NPP_WriteReady(), and
 *	ignore them in NPP_Write().  For a NP_ASFILE stream, they are
 *	still called but can safely be ignored using this strategy.
 */

int32 
NPP_WriteReady(NPP instance, NPStream *stream)
{
    
	if (instance != NULL)
		This = (PluginInstance*) instance->pdata;

  if (local_flag)
    {
      decode_data();
      return 0;
    }
  if (!required) {
    decode_data();
    return 0;
  }

  return required;
}


int32 
NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer)
{
	int i;
	char *curptr;
	
	if (instance != NULL)
	  {      
		This = (PluginInstance*) instance->pdata;
          }
	
	(int)curptr=(int)tempbuf+curwrt;	
        loaded=loaded+len;
	required=required-len; /* decrease required amount of data */
	curwrt=curwrt+len;
	memcpy(curptr, buffer, len);
	if (required==0)
	{
	  curwrt=0;	
	  memcpy(spoiler+(SLOTS-bank)*VIDEOUNIT, tempbuf, bank*VIDEOUNIT);
	  bank=0;
	}
        return len;		/* The number of bytes accepted */
}
 

NPError 
NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*) instance->pdata;

	playmode=STOP;
	started=0;
	SetMessage("Video sequence finished. You can press RePlay button to repeat...");
	XtSetSensitive(button3, True);
	return NPERR_NO_ERROR;
}

void 
NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
    return;
}
 
void 
NPP_Print(NPP instance, NPPrint* printInfo)
{
    return;
}

void 
NPP_URLNotify(NPP instance, const char *url, NPReason reason, void* notifyData)
{
    return;
}


/* **************** Here ends code for plugin in Netscape********** */

int decode_data()
{
    int i, result;

    if (playmode==PLAY)
 {
   if (first)
    {
      if (dump_flag)
      {
        outfile=open("video.263", O_RDWR|O_CREAT);
        if (outfile<0) printf("Error opening output file!\n");
      }

      if (local_flag)
      {
        base.infile=open(inputname, O_RDONLY);
	free(inputname);
        if (base.infile<0)
	    printf("Error opening input file: %d \n", errno);
        if (base.infile!=0)
          lseek(base.infile,0l,0);
	ld=&base;
	
        loaded=read(ld->infile, spoiler, SLOTS*VIDEOUNIT);
        if (loaded!=SLOTS*VIDEOUNIT)
          printf("Error reading from input file!\n");
	bank=0;
      }
       
     /* pointer to name of output files */

#if (defined DISPLAY || defined WINDOWS)
  if (outtype==T_X11 || outtype == T_WIN)
    outputname = "";
#endif

      ld = &base;
      initbits();         

#ifdef USE_TIME
  /* default is 99, then read frame rate from bitstream */
  framerate=10;
#endif

    framenum = 0;
    temp_ref = 0;
    prev_temp_ref -1; 
    } /* first */

 while(getheader()) 
  {
        if (first) {
	  if ((result=initdecoder())!=0)
	    {
	      fprintf(stdout, "Can not initialize decoder. Exiting...\n");
	      playmode=WAIT;
	      return;
	    }

#ifdef USE_TIME
#ifndef WIN32
        gettimeofday(&tstart,(struct timezone *)NULL);
        if (framerate > 0)
          gettimeofday(&tftarget,(struct timezone *)NULL);
#else
        startTime = timeGetTime();
        if (framerate > 0)
          targetTime = timeGetTime();
#endif
#endif
        first = 0;
        } /* first=1 */

    getpicture(&framenum);
    framenum++;
    
        if (bank!=0) /*prepare new data do decode*/
	  {
   	    curwrt=0;
	    for (i=0; i<(SLOTS-bank); i++)
	    {
	      memcpy(spoiler+i*VIDEOUNIT, spoiler+(bank+i)*VIDEOUNIT,VIDEOUNIT);/*shift*/
	    }
	    required=bank*VIDEOUNIT;
            if (!local_flag)
	      {		
	        return;
              }
	      else
	      {
                loaded+=read(ld->infile, spoiler+(SLOTS-bank)*VIDEOUNIT, required);
                bank=0;
	        return;		/* for X event handling */
	      }
	  }  

  } /* while(getheader()) */

  if (local_flag)
  {
    close(ld->infile);
    oldstrumien=malloc(sizeof(NPStream));
    memcpy(oldstrumien, strumien, sizeof(NPStream));
    NPP_DestroyStream(plugin, oldstrumien, 0);
  }
   
#ifdef USE_TIME
#ifndef WIN32
  gettimeofday(&tstop,(struct timezone *)NULL);
  runtime = 100*(tstop.tv_sec-tstart.tv_sec)
    + (tstop.tv_usec-tstart.tv_usec)/10000;
#else
  stopTime = timeGetTime();
  runtime = (stopTime - startTime) / 10;
#endif
  if (!quiet && runtime!=0)
    printf("%d.%02d seconds, %d frames, %d.%02d fps\n",
           runtime/100, runtime%100,
           framenum, ((10000*framenum+runtime/2)/runtime)/100,
           ((10000*framenum+runtime/2)/runtime)%100);
#endif

 }/* if playmode==PLAY */
}

static int initdecoder()
{
  int i, cc, size, result=0;
  FILE *cleared;

  /* clip table */
  if (!(clp=(unsigned char *)malloc(1024)))
    error("malloc failed\n");

  clp += 384;

  for (i=-384; i<640; i++)
    clp[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

  /* MPEG-1 = TMN parameters */
  matrix_coefficients = 5;


  switch (source_format) {
    case (SF_SQCIF):
      horizontal_size = 128;
      vertical_size = 96;
      break;
    case (SF_QCIF):
      horizontal_size = 176;
      vertical_size = 144;
      break;
    case (SF_CIF):
      horizontal_size = 352;
      vertical_size = 288;
      break;
    default:
      printf("ERROR: Illegal input format\n");
      exit(-1);
      break;
  }


  mb_width = horizontal_size/16;
  mb_height = vertical_size/16;
  coded_picture_width = horizontal_size;
  coded_picture_height = vertical_size;
  chrom_width =  coded_picture_width>>1;
  chrom_height = coded_picture_height>>1;
  blk_cnt = 6;

  for (cc=0; cc<3; cc++) {
    if (cc==0)
      size = coded_picture_width*coded_picture_height;
    else
      size = chrom_width*chrom_height;

    if (!(refframe[cc] = (unsigned char *)malloc(size)))
      error("malloc failed\n");

    if (!(oldrefframe[cc] = (unsigned char *)malloc(size)))
      error("malloc failed\n");

    if (!(bframe[cc] = (unsigned char *)malloc(size)))
      error("malloc failed\n");
  }

  for (cc=0; cc<3; cc++) {
    if (cc==0) {
      size = (coded_picture_width+64)*(coded_picture_height+64);
      if (!(edgeframeorig[cc] = (unsigned char *)malloc(size)))
        error("malloc failed\n");
      edgeframe[cc] = edgeframeorig[cc] + (coded_picture_width+64) * 32 + 32;
    }
    else {
      size = (chrom_width+32)*(chrom_height+32);
      if (!(edgeframeorig[cc] = (unsigned char *)malloc(size)))
        error("malloc failed\n");
      edgeframe[cc] = edgeframeorig[cc] + (chrom_width+32) * 16 + 16;
    }
  }

  if (expand) {
    for (cc=0; cc<3; cc++) {
      if (cc==0)
        size = coded_picture_width*coded_picture_height*4;
      else
        size = chrom_width*chrom_height*4;
      
      if (!(exnewframe[cc] = (unsigned char *)malloc(size)))
        error("malloc failed\n");
    }
  }

  /* Clear output file for concatenated storing */
  if (outtype == T_YUV_CONC) {
    if ((cleared = fopen(outputname,"wb")) == NULL) 
      error("couldn't clear outputfile\n");
    else
      fclose(cleared);
  }
  /* IDCT */
  if (refidct)
    init_idctref();
  else
    init_idct();

#ifdef DISPLAY
  if (outtype==T_X11) {
   result= init_display("");
  }
#endif
#ifdef WINDOWS
  if (outtype==T_WIN) {
    result=initDisplay(coded_picture_width, coded_picture_height);
  }
#endif

  return result;
}

void error(text)
char *text;
{
  fprintf(stderr,text);
  exit(1);
}

/* trace output */
void printbits(code,bits,len)
int code,bits,len;
{
  int i;
  for (i=0; i<len; i++)
    printf("%d",(code>>(bits-1-i))&1);
}


static int getval(argv)
char *argv[];
{
  int val;

  if (sscanf(argv[1]+2,"%d",&val)!=1)
    return 0;

  while (isdigit(argv[1][2]))
    argv[1]++;

  return val;
}

#ifdef USE_TIME
#ifndef WINDOWS

/* Unix version */
void doframerate(int pb)
{
  struct timeval tfdiff;
  const float REF_FRAME_RATE = 29.97;

  /* Compute desired frame rate */
  if (framerate <= 0)
    return;
  
  if (framerate != 99) {
    tftarget.tv_usec += 1000000 / framerate;
  }
  else {
    if (pb) {
      tftarget.tv_usec += 1000000 / (REF_FRAME_RATE/trb);
    }
    else {
      tftarget.tv_usec += 1000000 / (REF_FRAME_RATE/(trd-trb));
    }
  }
  
  /* update target time by pause time */
  tftarget.tv_usec += ptime.tv_usec;
  tftarget.tv_sec+=ptime.tv_sec;
  ptime.tv_usec=pstop.tv_usec=pstart.tv_usec=0;
  ptime.tv_sec=pstop.tv_sec=pstart.tv_sec=0;
    
  /* this is where we should be */
  if (tftarget.tv_usec >= 1000000)
  {
    tftarget.tv_usec -= 1000000;
    tftarget.tv_sec++;
  }

  /* this is where we are */
  gettimeofday(&tfdiff,(struct timezone *)NULL);

  tfdiff.tv_usec = tftarget.tv_usec - tfdiff.tv_usec;
  tfdiff.tv_sec  = tftarget.tv_sec  - tfdiff.tv_sec;
  if (tfdiff.tv_usec < 0)
  {
    tfdiff.tv_usec += 1000000;
    tfdiff.tv_sec--;
  }

  /* See if we are already lagging behind */
  if (tfdiff.tv_sec < 0 || (tfdiff.tv_sec == 0 && tfdiff.tv_usec <= 0))
    return;
  
  /* Spin for awhile */
  select(0,NULL,NULL,NULL,&tfdiff);
}

#else 

/* Win32 version */
void doframerate(int pb)
{
  DWORD currentTime;
  int diffTime;
  const float REF_FRAME_RATE = (float)29.97;

  /* Compute desired frame rate */
  if (framerate <= 0)
    return;
  
  if (framerate != 99) {
    targetTime += 1000 / framerate;
  }
  else {
    if (pb) {
      targetTime += (int)(1000 / (REF_FRAME_RATE/trb));
    }
    else {
      targetTime += (int)(1000 / (REF_FRAME_RATE/(trd-trb)));
    }
  }

  /* this is where we are */
  currentTime = timeGetTime();
  
  diffTime = targetTime - currentTime;

  /* See if we are already lagging behind */
  if (diffTime <= 0)
    return;
  
  /* Spin for awhile */
  Sleep(diffTime);   
     /* this is not a very accurate timer */
}

#endif

#endif

void start(w, client_data, cbs)
Widget w;
XtPointer client_data;
XmPushButtonCallbackStruct *cbs;
{
  if ((playmode!=STOP)&&(playmode!=PLAY))
    {
      started=1;
      if (playmode==PAUSE) /* pause time calculation for frame rate */
      {
	gettimeofday(&pstop,(struct timezone *)NULL);
	ptime.tv_sec=pstop.tv_sec-pstart.tv_sec;
	if ((ptime.tv_usec=pstop.tv_usec-pstart.tv_usec)<0)
	{
	  ptime.tv_usec+=1000000;
	  ptime.tv_sec--;
	}
      }
      SetMessage("Video playing...");
      playmode=PLAY;
      decode_data();
    } 
}

void my_pause(void)
{
  if (playmode==PLAY)
    {
      SetMessage("Video paused. Press PLAY to continue...");
      gettimeofday(&pstart,(struct timezone *)NULL);
      playmode=PAUSE;
      decode_data();
    }
}

void replay(w, client_data, cbs)
Widget w;
XtPointer client_data;
XmPushButtonCallbackStruct *cbs;
{
  if (playmode==STOP)
    {
      XtSetSensitive(button3, False);
      SetMessage("Video re-playing...");
      playmode=PLAY;
      started=1;
      first=1;
      local_flag=0;
      required=0;
      if (newcmap) XUninstallColormap(display, cmap);
      NPN_GetURL(plugin, strumien->url, 0L);
    }
}

void toggled(w, client_data, cbs)
Widget w;
XtPointer client_data;
XmToggleButtonCallbackStruct *cbs;
{
    XmString text;
  
  if (!started)
    {
      if (dump_flag)
      {
        dump_flag = 0;
	text = XmStringCreateSimple("Do not dump to file");
      }
      else
	{
	  dump_flag = 1;
	  text = XmStringCreateSimple("Dump to video.263");
	}
      XtVaSetValues(tbutton, XmNlabelString, text);
      XmStringFree(text);      
    }
  else XmToggleButtonSetState(tbutton, !cbs->set, False);
}

void toggformat(w, client_data, cbs)
Widget w;
XtPointer client_data;
XmToggleButtonCallbackStruct *cbs;
{
    XmString text;
  
  if (!started)
    {
      if (expand)
      {
        expand = 0;
	text = XmStringCreateSimple("Do not expand");
      }
      else
	{
	  expand = 1;
	  text = XmStringCreateSimple("Expand to CIF");
	}
      XtVaSetValues(tformat, XmNlabelString, text);
      XmStringFree(text);      
    }
  else XmToggleButtonSetState(tformat, !cbs->set, False);
}

void SetMessage(const char *msg)
{
  XmString str=XmStringCreateSimple((char*)msg);
  XtVaSetValues(messagelabel,
		XmNlabelString, str,
		NULL);
  XmStringFree(str);
}

/* DISPLAY */

#ifdef SH_MEM
static int HandleXError(dpy, event)
Display *dpy;
XErrorEvent *event;
{
  gXErrorFlag = 1;

  return 0;
}

static void InstallXErrorHandler()
{
  XSetErrorHandler(HandleXError);
  XFlush(display);
}

static void DeInstallXErrorHandler()
{
  XSetErrorHandler(NULL);
  XFlush(display);
}

#endif

/* disconnect from server, free all resources */

void close_display()
{
    XCloseDisplay(display);
    return;
}

/* connect to server, create and map window,
 * allocate colors and (shared) memory
 */
int init_display(name)
char *name;
{
  int crv, cbu, cgu, cgv;
  int y, u, v, r, g, b;
  int i, j;
  char dummy;
  int dpy_class;
  int private;
  XColor xcolor;
  int screen;
  char *hello = "Video!";
  XSizeHints hint;
  XSetWindowAttributes xswa;
  unsigned long tmp_pixel;
  double min, temp;
  unsigned int mask;
  int index, num;
  Position t_y;
  Colormap *maps;

  if (display == NULL)
    error("Can not open display\n");

  screen = DefaultScreen(display);

  dpy_class = visual->class;

  if (!((dpy_class == TrueColor && dpy_depth == 32)
	|| (dpy_class == TrueColor && dpy_depth == 24)
	|| (dpy_class == TrueColor && dpy_depth == 16)
        || (dpy_class == PseudoColor && dpy_depth == 8)))
    printf ("requires 8 bit PseudoColor or 16/24/32 bit TrueColor display\n");

  /* width and height of the display window */
  if (expand) {
    hint.min_width = hint.width = 2*horizontal_size;
    hint.min_height = hint.height = 2*vertical_size;
  }
  else {
    hint.min_width = hint.width = horizontal_size;
    hint.min_height = hint.height = vertical_size;
  }
/*  hint.min_width = hint.width = horizontal_size; 
  hint.min_height = hint.height = vertical_size;
*/
  hint.max_width=700;
  hint.max_height=700;
  hint.flags = PSize | PMinSize | PMaxSize;
  
  /* Make the window */

  if (This->winPels > NW_width) 
  {
    NW_width += This->winPels-176;
    NW_height += This->winLines-144;
    XResizeWindow(display, okno_from_netscape, NW_width, NW_height);
    XtVaGetValues(button1, XmNy, &t_y, NULL);
    XtRealizeWidget(This->netscape_widget);
  }

  mask = CWBackPixel | CWBorderPixel | CWBackingStore;
  if (dpy_depth == 32 || dpy_depth == 24 || dpy_depth == 16) {
    mask |= CWColormap;
    xswa.colormap = XCreateColormap (display, DefaultRootWindow(display),
				     visual, AllocNone);
  }

  xswa.background_pixel = bg;
  xswa.border_pixel = fg;
  xswa.backing_store = Always;

  hint.x=50; /* location within browser's window */
  hint.y=This->buttonsLine+60;
  
  window = XCreateWindow (display, okno_from_netscape,
			  hint.x, hint.y, hint.width, hint.height,
			  0, dpy_depth, InputOutput, visual, mask
			  , &xswa);
  if (window==0)
  {
    fprintf(stdout, "ERROR: Display window can not be created!!!\n");
    fflush(stdout);
  }
  
  XSelectInput(display, window, StructureNotifyMask);

  /* Tell other applications about this window */

  XSetStandardProperties (display, window, hello, hello, None, NULL, 0, &hint);

  /* Map window. */

  XMapWindow(display, window);

  /* Wait for map. */
  do
  {
    XNextEvent(display, &xev);
  }
  while (xev.type != MapNotify || xev.xmap.event != window);

  XSelectInput(display, window, NoEventMask);

  /* allocate colors */
  
  if (dpy_depth == 8) {
    XWindowAttributes xwa;

    XGetWindowAttributes(display, window, &xwa);
    cmap = XCreateColormap (display, window, xwa.visual, AllocNone);
    newcmap=1;
    XSetWindowColormap(display, window, cmap);
    XInstallColormap(display, cmap);
/*    maps=XListInstalledColormaps(display, window, &num);
    printf("Installed %d colormaps:\n", num);
    for (i=0;i<num;i++)
    {
      printf("%d\n", *maps);
      maps+=sizeof(Colormap);
    }
  */  
    private = 0;

    /* matrix coefficients */
    crv = convmat[matrix_coefficients][0];
    cbu = convmat[matrix_coefficients][1];
    cgu = convmat[matrix_coefficients][2];
    cgv = convmat[matrix_coefficients][3];

    /* color allocation:
     * i is the (internal) 8 bit color number, it consists of separate
     * bit fields for Y, U and V: i = (yyyyuuvv), we don't use yyyy=0000
     * yyyy=0001 and yyyy=1111, this leaves 48 colors for other applications
     *
     * the allocated colors correspond to the following Y, U and V values:
     * Y:   40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200, 216, 232
     * U,V: -48, -16, 16, 48
     *
     * U and V values span only about half the color space; this gives
     * usually much better quality, although highly saturated colors can
     * not be displayed properly
     *
     * translation to R,G,B is implicitly done by the color look-up table
     */

    for (i=32; i<240; i++) {
      /* color space conversion */
      y = 16*((i>>4)&15) + 8;
      u = 32*((i>>2)&3)  - 48;
      v = 32*(i&3)       - 48;

      y = 76309 * (y - 16); /* (255/219)*65536 */
      
      r = clp[(y + crv*v + 32768)>>16];
      g = clp[(y - cgu*u -cgv*v + 32768)>>16];
      b = clp[(y + cbu*u + 32786)>>16];

      /* X11 colors are 16 bit */
      xcolor.red   = r << 8;
      xcolor.green = g << 8;
      xcolor.blue  = b << 8;

      c[i-32]=xcolor; /* store color in table */

      
    if (XAllocColor(display, cmap, &xcolor) != 0)
      {
	pixel[i] = xcolor.pixel;
      }
      else 
      {
	if (!private)
	  printf("Missing colors - %d were available\n", i-32);
	private=1;
	min=1000000;
	index=i;
	for (j=0;j<(i-32);j++)
	{
	  temp = sqrt((xcolor.red-c[j].red)*(xcolor.red-c[j].red)
	             +(xcolor.green-c[j].green)*(xcolor.green-c[j].green)
		     +(xcolor.blue-c[j].blue)*(xcolor.blue-c[j].blue)
		     );
	  if (temp<min)
	  {
	    min=temp;
	    index=j;
	  }
	}
	pixel[i] = pixel[index]; 
	}
    }
  }

#ifdef SH_MEM
  if (XShmQueryExtension(display))
    shmem_flag = 1;
  else
  {
    shmem_flag = 0;
    if (!quiet)
      fprintf(stderr, "Shared memory not supported\nReverting to normal "
	      "Xlib\n");
  }

  if (shmem_flag)
    CompletionType = XShmGetEventBase(display) + ShmCompletion;

  InstallXErrorHandler();

  if (shmem_flag)
  {

    if (expand)
      ximage = XShmCreateImage(display, visual, dpy_depth, ZPixmap, NULL,
			       &shminfo1,
			       2*coded_picture_width, 2*coded_picture_height);
    else
      ximage = XShmCreateImage(display, visual, dpy_depth, ZPixmap, NULL,
			       &shminfo1,
			       coded_picture_width, coded_picture_height);
    

    /* If no go, then revert to normal Xlib calls. */

    if (ximage==NULL)
    {
      if (ximage!=NULL)
        XDestroyImage(ximage);
      if (!quiet)
        fprintf(stderr, "Shared memory error, disabling (Ximage error)\n");
      goto shmemerror;
    }

    /* Success here, continue. */

    shminfo1.shmid = shmget(IPC_PRIVATE, 
                            ximage->bytes_per_line * ximage->height,
                            IPC_CREAT | 0777);

    if (shminfo1.shmid<0)
    {
      XDestroyImage(ximage);
      if (!quiet)
        fprintf(stderr, "Shared memory error, disabling (seg id error)\n");
      goto shmemerror;
    }

    shminfo1.shmaddr = (char *) shmat(shminfo1.shmid, 0, 0);
    shminfo2.shmaddr = (char *) shmat(shminfo2.shmid, 0, 0);

    if (shminfo1.shmaddr==((char *) -1))
    {
      XDestroyImage(ximage);
      if (shminfo1.shmaddr!=((char *) -1))
        shmdt(shminfo1.shmaddr);
      if (!quiet)
      {
        fprintf(stderr, "Shared memory error, disabling (address error)\n");
      }
      goto shmemerror;
    }

    ximage->data = shminfo1.shmaddr;
    dithered_image = (unsigned char *)ximage->data;
    shminfo1.readOnly = False;
    XShmAttach(display, &shminfo1);

    XSync(display, False);

    if (gXErrorFlag)
    {
      /* Ultimate failure here. */
      XDestroyImage(ximage);
      shmdt(shminfo1.shmaddr);
      if (!quiet)
        fprintf(stderr, "Shared memory error, disabling.\n");
      gXErrorFlag = 0;
      goto shmemerror;
    }
    else
    {
      shmctl(shminfo1.shmid, IPC_RMID, 0);
    }

    if (!quiet)
    {
      fprintf(stderr, "Sharing memory.\n");
    }
  }
  else
  {
shmemerror:
    shmem_flag = 0;
#endif

    if (expand) {
      ximage = XCreateImage(display,visual,dpy_depth,ZPixmap,0,&dummy,
			    2*coded_picture_width,2*coded_picture_height,8,0);
      if (!(dithered_image =
	    (unsigned char *)malloc(coded_picture_width*coded_picture_height*
				    (dpy_depth > 8 ? sizeof (int)*4 : 
				     sizeof (unsigned char))*4)))
	error("malloc failed");
    }
    else {
      ximage = XCreateImage(display,visual,dpy_depth,ZPixmap,0,&dummy,
			    coded_picture_width,coded_picture_height,8,0);
      if (!(dithered_image =
	    (unsigned char *)malloc(coded_picture_width*coded_picture_height*
				    (dpy_depth > 8 ? sizeof (int) : 
				     sizeof (unsigned char)))))
	error("malloc failed");
    }

#ifdef SH_MEM
  }

  DeInstallXErrorHandler();
#endif


  if (dpy_depth == 32 || dpy_depth == 24 || dpy_depth == 16) {
    XWindowAttributes xwa;

    XGetWindowAttributes(display, window, &xwa);


    wpixel[0] = xwa.visual->red_mask;
    wpixel[1] = xwa.visual->green_mask;
    wpixel[2] = xwa.visual->blue_mask;

    /* If the colors in 16/24/32-bit mode are wrong, try this instead
       of the above three lines */
    /*
    wpixel[2] = xwa.visual->red_mask;
    wpixel[1] = xwa.visual->green_mask;
    wpixel[0] = xwa.visual->blue_mask;
    */

    InitColorDither(dpy_depth == 24 || dpy_depth == 32);
  }
  else {
    ord4x4_dither_init ();
  }
      SetMessage("Video playing...");
}


void exit_display()
{
#ifdef SH_MEM
  if (shmem_flag)
  {
    XShmDetach(display, &shminfo1);
    XDestroyImage(ximage);
    shmdt(shminfo1.shmaddr);
  }
#endif
}

static void display_image(ximage,dithered_image)
XImage *ximage;
unsigned char *dithered_image;
{
  int t = 1;

 /* Always work in native bit and byte order. This tells Xlib to
    reverse bit and byte order if necessary when crossing a
    network. Frankly, this part of XImages is somewhat
    underdocumented, so this may not be exactly correct.  */

  if (*(char *)&t == 1) {
    ximage->byte_order = LSBFirst;
    ximage->bitmap_bit_order = LSBFirst;
  }
  else {
    ximage->byte_order = MSBFirst;
    ximage->bitmap_bit_order = MSBFirst;
  }    

  /* display dithered image */
#ifdef SH_MEM
  if (shmem_flag)
  {
    XShmPutImage(display, window, gc, ximage, 
       	         0, 0, 0, 0, ximage->width, ximage->height, True);
    XFlush(display);
      
    while (1)
    {
      XEvent xev;
	
      XNextEvent(display, &xev);
      if (xev.type == CompletionType)
        break;
    }
  }
  else 
#endif
  {
    ximage->data = (char *) dithered_image;
    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, ximage->width, ximage->height);
  }
}

void dither(src)
unsigned char *src[];
{
  if (dpy_depth == 24 || dpy_depth == 32) {
    if (ximage->bits_per_pixel == 24)
      ConvertYUVtoRGB(src[0],src[1],src[2], dithered_image,
		      coded_picture_width,
		      coded_picture_height);
    else
      Color32DitherImage(src, dithered_image);
  }  
  else if (dpy_depth == 16) {
    Color16DitherImage(src, dithered_image);
  }
  else {
    ord4x4_dither_frame (src, dithered_image);
  }

  display_image (ximage, dithered_image);
}


/* DISPLAY */
int finishdecoder()
{

  free(spoiler);
  free(tempbuf);
  free(strumien);

#ifdef USE_TIME
  gettimeofday(&tstop,(struct timezone *)NULL);
  runtime = 100*(tstop.tv_sec-tstart.tv_sec)
    + (tstop.tv_usec-tstart.tv_usec)/10000;
  if (!quiet && runtime!=0)
    printf("%d.%02d seconds, %d frames, %d.%02d fps\n",
	   runtime/100, runtime%100,
	   framenum, ((10000*framenum+runtime/2)/runtime)/100,
	   ((10000*framenum+runtime/2)/runtime)%100);
	   
#endif

#ifdef DISPLAY
  if (outtype==T_X11)
    exit_display();
#endif
#ifdef WINDOWS
  if (outtype == T_WIN)
    closeDisplay();
#endif

  return 0;
}


void native_Pl_aa(JRIEnv* env, Pl* self, int liczba)
{
   if (liczba)
     {
       timerId=XtAppAddTimeOut(This->appContext,1,start,NULL);
     }
   else 
     {
       timerId=XtAppAddTimeOut(This->appContext,1,my_pause,NULL);
     }
   return;
}

